import os
from os import listdir
from os.path import isfile, join

import yaml

from ..core.task import Flow
from . import FlowStore


class ParadeFlowStore(FlowStore):
    """
    This is *JUST A PLACEHOLDER* implementation
    """

    def load(self, name):
        flow_file = os.path.join(self.flow_dir, name + '.yml')
        if os.path.exists(flow_file):
            with open(flow_file, 'r') as f:
                content = f.read()
                rawflow = yaml.load(content)
                _deps = {}
                for dep in rawflow['deps']:
                    (l, r) = tuple(dep.split('->'))
                    _deps[l] = set(r.split(','))
                flow = Flow(name, rawflow['tasks'], _deps)
                return flow
        return None

    flow_dir = None

    def initialize(self, context, conf):
        FlowStore.initialize(self, context, conf)
        self.flow_dir = join(self.context.workdir, 'flows')

    def list(self):
        return set([os.path.splitext(file)[0]
                    for file in listdir(self.flow_dir) if isfile(join(self.flow_dir, file)) and file.endswith('.yml')])

    def create(self, flow, *tasks, deps=None):
        if not deps:
            deps = {}

        Flow(flow, tasks, deps)

        dep_lines = list(map(lambda x: x[0] + '->' + ','.join(x[1]), deps.items()))
        create_flow = {
            'tasks': list(tasks),
            'deps': dep_lines
        }

        class IndentDumper(yaml.Dumper):
            def increase_indent(self, flow=False, indentless=False):
                return super(IndentDumper, self).increase_indent(flow, False)

        os.makedirs(self.flow_dir, exist_ok=True)
        flow_file = os.path.join(self.flow_dir, flow + '.yml')
        with open(flow_file, 'w') as f:
            yaml.dump(create_flow, f, Dumper=IndentDumper, default_flow_style=False)

    def delete(self, flow):
        flow_file = os.path.join(self.flow_dir, flow + '.yml')
        if os.path.exists(flow_file):
            os.remove(flow_file)



from .. import ParadeCommand


class RemoveFlowCommand(ParadeCommand):
    requires_workspace = True

    @property
    def aliases(self):
        return ['remove', 'delete', 'del']

    def run_internal(self, context, **kwargs):
        flow = kwargs.get('flow')
        flowstore = context.get_flowstore()
        flowstore.delete(flow)

    def short_desc(self):
        return 'delete the specified flow'

    def config_parser(self, parser):
        parser.add_argument('flow', help='the flow to delete')



import datetime
import time

import pytz
from dateutil import relativedelta


def datetime_str_to_timestamp(datetime_str, tz=None):
    dt = datetime.datetime.strptime(datetime_str, '%Y-%m-%d %H:%M:%S')
    return datetime_to_timestamp(dt, tz)


def datetime_to_timestamp(dt, tz=None):
    if tz:
        dt = pytz.timezone(tz).localize(dt)
        return int((dt - datetime.datetime(1970, 1, 1, tzinfo=pytz.utc)).total_seconds())
    return int(time.mktime(dt.timetuple()))


def timestamp_to_datetime(ts, tz=None):
    if tz:
        return datetime.datetime.fromtimestamp(ts, pytz.timezone(tz))
    return datetime.datetime.fromtimestamp(ts)


def gen_time_df(periods, unit='D', start=None, complete=False):
    import pandas as pd

    if not start:
        today = datetime.date.today()
        if unit == 'D':
            start = today - datetime.timedelta(days=periods - 1)
        elif unit == 'W':
            start = today - datetime.timedelta(days=today.weekday()) - datetime.timedelta(weeks=periods - 1)
        elif unit == 'M':
            start = today.replace(day=1) - relativedelta.relativedelta(months=periods - 1)

    if unit == 'D':
        df = pd.DataFrame([[x]
                           for x in pd.date_range(start, periods=periods, freq='D')],
                          columns=['day'])
        if complete:
            df['weekday'] = df['day'].dt.dayofweek
            df['week_start'] = df.apply(lambda x: x['day'] - datetime.timedelta(days=x['weekday']), axis=1)
            df['week_end'] = df.apply(lambda x: x['week_start'] + datetime.timedelta(days=6), axis=1)
            df['month_start'] = df.apply(lambda x: x['day'].replace(day=1), axis=1)
            df['month_end'] = df.apply(
                lambda x: x['day'].replace(day=1) + relativedelta.relativedelta(months=1) - datetime.timedelta(days=1),
                axis=1)
            df['month'] = df['day'].dt.month
            df['year'] = df['day'].dt.year
            df['month'] = df.apply(lambda r: int(str(r['year']) + str(r['month']).zfill(2)), axis=1)
            df.drop('year', axis=1, inplace=True)

    elif unit == 'W':
        df = pd.DataFrame([[x]
                           for x in pd.date_range(start, periods=periods, freq='W-MON')],
                          columns=['week_start'])
        df['week_end'] = df.apply(lambda x: x['week_start'] + datetime.timedelta(days=6), axis=1)
        if complete:
            df['month_start'] = df.apply(lambda x: x['week_start'].replace(day=1), axis=1)
            df['month_end'] = df.apply(
                lambda x: x['week_start'].replace(day=1) + relativedelta.relativedelta(months=1) - datetime.timedelta(
                    days=1), axis=1)
            df['month'] = df['week_start'].dt.month
            df['year'] = df['week_start'].dt.year
            df['month'] = df.apply(lambda r: int(str(r['year']) + str(r['month']).zfill(2)), axis=1)
            df.drop('year', axis=1, inplace=True)

    elif unit == 'M':
        df = pd.DataFrame([[x]
                           for x in pd.date_range(start, periods=periods, freq='MS')],
                          columns=['month_start'])
        df['month_end'] = df.apply(
            lambda x: x['month_start'].replace(day=1) + relativedelta.relativedelta(months=1) - datetime.timedelta(
                days=1), axis=1)
        df['month'] = df['month_start'].dt.month
        df['year'] = df['month_start'].dt.year
        df['month'] = df.apply(lambda r: int(str(r['year']) + str(r['month']).zfill(2)), axis=1)
        df.drop('year', axis=1, inplace=True)

    else:
        raise RuntimeError

    return df



from collections import defaultdict

from .recorder import ParadeRecorder
from ..config import ConfigStore, ConfigEntry
from ..connection import Connection
from ..core.task import Task
from ..error.task_errors import DuplicatedTaskExistError
from ..flowrunner import FlowRunner
from ..flowstore import FlowStore
from ..notify import Notifier
from ..utils.log import parade_logger as logger
from ..utils.modutils import get_class, iter_classes


class Context(object):
    """
    The executor context to support the ETL job executed by the engine
    """

    def __init__(self, bootstrap):
        workspace_settings = bootstrap['workspace']

        self.name = workspace_settings['name']
        self.workdir = workspace_settings['path']

        self._task_dict = defaultdict(Task)
        self._conn_cache = defaultdict(Connection)
        self._notifier = None
        self._flowstore = None
        self._flowrunner = None
        self._configstore = None

        config_settings = bootstrap['config']
        self.conf = ConfigEntry({'config': config_settings})

        self._init_configstore()

        self.conf = self._configstore.load()

    # ========================Context as a task store=========================
    # ========================================================================
    def load_tasks(self, name=None, task_class=Task):
        """
        generate the task dict [task_key => task_obj]
        :return:
        """
        d = {}
        for task_class in iter_classes(task_class, self.name + '.task'):
            task = task_class()
            task_name = task.name
            if name and task_name != name:
                continue
            if task_name in d:
                raise DuplicatedTaskExistError(task=task_name)
            d[task_name] = task
        return d

    def get_task(self, name, task_class=Task):
        if name in self._task_dict:
            return self._task_dict[name]
        task = self.load_tasks(name, task_class=task_class)[name]
        self._task_dict[name] = task
        return task

    def list_tasks(self, task_class=Task):
        """
        get the task namelist
        :return:
        """
        return self.load_tasks(task_class=task_class).keys()

    # ========================================================================

    # ====================Context as a connection store=======================
    # ========================================================================
    def get_connection(self, conn_key):
        """
        Get the connection with the connection key
        :param conn_key: the key of the connection
        :return: the connection instance
        """

        if conn_key not in self._conn_cache:
            conn_conf = self.conf['connection']
            assert conn_conf.has(conn_key), 'connection {} is not configured'.format(conn_key)
            self._conn_cache[conn_key] = self._load_plugin('connection', Connection, plugin_key=conn_key)

        return self._conn_cache[conn_key]

    def load(self, table, conn=None, **kwargs):
        if not conn:
            raise ValueError('connection not provided')
        return self.get_connection(conn).load(table, **kwargs)

    def load_query(self, query, conn=None, **kwargs):
        if not conn:
            raise ValueError('connection not provided')
        return self.get_connection(conn).load_query(query, **kwargs)

    def store(self, df, table, conn=None, **kwargs):
        if not conn:
            raise ValueError('connection not provided')
        return self.get_connection(conn).store(df, table, **kwargs)

    # ========================================================================

    # ====================Context as other plugin store=======================
    # ========================================================================
    def get_flowstore(self):
        """
        Get the flow store
        :return: the flow store instance
        """

        if not self._flowstore:
            self._flowstore = self._load_plugin('flowstore', FlowStore)

        return self._flowstore

    def get_flowrunner(self):
        """
        Get the flow store
        :return: the flow store instance
        """

        if not self._flowrunner:
            self._flowrunner = self._load_plugin('flowrunner', FlowRunner)

        return self._flowrunner

    def _init_configstore(self):
        if not self._configstore:
            self._configstore = self._load_plugin('config', ConfigStore)

    def _load_plugin(self, plugin_token, plugin_class, plugin_key=None, default_conf=None):
        conf = default_conf

        if self.conf.has(plugin_token):
            conf = self.conf[plugin_token][plugin_key] if plugin_key else self.conf[plugin_token]
            assert conf.has('driver'), 'no driver provided in {} section'.format(plugin_token)
            driver = conf['driver']
        else:
            driver = 'default'

        plugin_cls = get_class(driver, plugin_class, 'parade.' + plugin_token, self.name + '.contrib.' + plugin_token)
        assert plugin_cls
        plugin = plugin_cls()
        plugin.initialize(self, conf)
        return plugin

    # ========================================================================

    @property
    def sys_recorder(self):
        """
        Get the parade system connection
        :return:
        """
        if not self.conf.has('checkpoint.connection'):
            return None
        checkpoint_conn_key = self.conf['checkpoint.connection']
        conn = self.get_connection(checkpoint_conn_key)

        try:
            return ParadeRecorder(self.name, conn)
        except:
            logger.warn('Parade recorder initialized failed!')
            return None

    def on_flow_start(self, flow):
        # TODO maybe we should set the flow state to `pending` here
        self.sys_recorder.init_record_tables()
        return self.sys_recorder.create_flow_record(flow.name, flow.tasks)

    def on_flow_success(self, flow_id):
        self.sys_recorder.mark_flow_success(flow_id)

    def on_flow_failed(self, flow_id):
        self.sys_recorder.mark_flow_failed(flow_id)

    def on_task_pending(self, task, checkpoint, flow_id, flow):
        self.sys_recorder.init_record_tables()
        return self.sys_recorder.create_task_record(task, checkpoint, flow_id, flow)

    def on_task_cancelled(self, task, failed_deps):
        self.sys_recorder.mark_task_cancelled(task.exec_id, failed_deps)

        if hasattr(self, 'webapp'):
            socketio = self.webapp.extensions['socketio']
            socketio.emit('update', {
                'flow-id': task.flow_id,
                'task': task.name,
                'task-id': task.exec_id,
                'event': 'task-cancelled'
            }, namespace='/exec', room=str(task.flow_id))

    def on_task_start(self, task):
        self.sys_recorder.mark_task_start(task.exec_id)

        if hasattr(self, 'webapp'):
            socketio = self.webapp.extensions['socketio']
            socketio.emit('update', {
                'flow-id': task.flow_id,
                'task': task.name,
                'task-id': task.exec_id,
                'event': 'task-started'
            }, namespace='/exec', room=str(task.flow_id))

    def on_task_success(self, task):
        self.sys_recorder.mark_task_success(task.exec_id)

        if hasattr(self, 'webapp'):
            socketio = self.webapp.extensions['socketio']
            socketio.emit('update', {
                'flow-id': task.flow_id,
                'task': task.name,
                'task-id': task.exec_id,
                'event': 'task-succeeded'
            }, namespace='/exec', room=str(task.flow_id))

        if task.notify_success and self.get_notifier() is not None:
            self.get_notifier().notify_success(self.name)

    def on_task_failed(self, task, err):
        self.sys_recorder.mark_task_failed(task.exec_id, err)

        if hasattr(self, 'webapp'):
            socketio = self.webapp.extensions['socketio']
            socketio.emit('update', {
                'flow-id': task.flow_id,
                'task': task.name,
                'task-id': task.exec_id,
                'event': 'task-failed'
            }, namespace='/exec', room=str(task.flow_id))

        if task.notify_fail and self.get_notifier() is not None:
            self.get_notifier().notify_error(self.name, str(err))

    def get_notifier(self):
        """
        Get the notifier with the notify key
        :return: the notifier instance
        """

        if not self._notifier:
            self._notifier = self._load_plugin('notify', Notifier)

        return self._notifier



from io import BytesIO

from ..connection import Connection
import pandas as pd
import os
import datetime
from pandas.compat import cPickle as pkl


class LocalFile(Connection):
    @staticmethod
    def export(df, table, export_type='excel', target_path=None, if_exists='replace', suffix=None):
        if target_path:
            if export_type == 'excel' and not suffix:
                suffix = 'xlsx'
            target_file = os.path.join(target_path,
                                       table + '-' + str(datetime.date.today())) + '.' + str(suffix or export_type)
            if if_exists == 'replace' and os.path.exists(target_file):
                os.remove(target_file)
            export_io = target_file

        else:
            export_io = BytesIO()

        if export_type == 'excel':
            writer = pd.ExcelWriter(export_io, engine='xlsxwriter')
            df.to_excel(writer, index=False)
            writer.save()
        elif export_type == 'csv':
            export_io = BytesIO(df.to_csv(target_path, index=False, chunksize=4096).encode())
        elif export_type == 'json':
            export_io = BytesIO(df.to_json(target_path, orient='records').encode())
        elif export_type == 'pickle':
            pkl.dump(df, export_io, protocol=pkl.HIGHEST_PROTOCOL)
        else:
            raise NotImplementedError("export type {} is not supported".format(export_type))
        return export_io, table + '.' + str(suffix or export_type)

    def store(self, df, table, **kwargs):
        target_path = self.datasource.db if self.datasource.db else '.'
        export_type = self.datasource.protocol if self.datasource.protocol else 'json'
        if_exists = kwargs.get('if_exists', 'replace')

        os.makedirs(target_path, exist_ok=True)
        LocalFile.export(df, table, export_type=export_type, target_path=target_path, if_exists=if_exists)

    def load_query(self, query, **kwargs):
        raise NotImplementedError

    def load(self, table, **kwargs):
        raise NotImplementedError



import argparse
import sys

from parade.command import ParadeCommand
from parade.config import ConfigStore
from parade.core.context import Context
from parade.utils.modutils import iter_classes
from parade.utils.workspace import inside_workspace, load_bootstrap


def _get_commands(in_workspace):
    d = {}
    for cmd in iter_classes(ParadeCommand, 'parade.command', class_filter=lambda cls: cls != ParadeCommand):
        if in_workspace or not cmd.requires_workspace:
            cmd_name = cmd.__module__.split('.')[-1]
            cmd_group = None if len(cmd.__module__) == len('parade.command') + 1 + len(cmd_name) else cmd.__module__[
                                                                                                      len(
                                                                                                          'parade.command') + 1:-len(
                                                                                                          cmd_name) - 1]
            if not cmd_group:
                d[cmd_name] = (cmd(), cmd_name, None)
            else:
                cmd_path = cmd_group + '.' + cmd_name
                d[cmd_path] = (cmd(), cmd_name, cmd_group)
                for cmd_alias in d[cmd_path][0].aliases:
                    d[cmd_group + '.' + cmd_alias] = d[cmd_path]

    return d


def _get_config_repo(driver, uri, workdir=None):
    for repo in iter_classes(ConfigStore, 'parade.config', class_filter=lambda cls: cls != ConfigStore):
        repo_name = repo.__module__.split('.')[-1]
        if repo_name == driver:
            return repo(uri, workdir=workdir)


def execute():
    # load the commands and parse arguments
    parser = argparse.ArgumentParser(description='The CLI of parade engine.')
    inworkspace = inside_workspace()
    cmd_parsers = parser.add_subparsers(dest='command')
    cmds = _get_commands(inworkspace)

    sub_cmd_parsers_dict = {}
    for cmd, cmd_name, cmd_group in cmds.values():
        if not cmd_group:
            cmd_parser = cmd_parsers.add_parser(cmd_name, help=cmd.help())
            cmd.config_parser(cmd_parser)
        else:
            if cmd_group in sub_cmd_parsers_dict:
                sub_cmd_parsers = sub_cmd_parsers_dict[cmd_group]
            else:
                group_parser = cmd_parsers.add_parser(cmd_group, help=(cmd_group + '-related sub commands'))
                sub_cmd_parsers = group_parser.add_subparsers(dest='subcommand')
                sub_cmd_parsers_dict[cmd_group] = sub_cmd_parsers
            sub_cmd_parser = sub_cmd_parsers.add_parser(cmd_name, aliases=cmd.aliases, help=cmd.help())
            cmd.config_parser(sub_cmd_parser)
    args = parser.parse_args()

    if not args.command:
        parser.print_usage()
        return 0

    cmd_path = args.command if not hasattr(args, 'subcommand') else args.command + '.' + args.subcommand

    command = cmds[cmd_path][0]
    command_args = args.__dict__

    context = None
    if inworkspace and command.requires_workspace:
        bootstrap = load_bootstrap()
        context = Context(bootstrap)

    return command.run(context, **command_args)


if __name__ == '__main__':
    retcode = execute()
    sys.exit(retcode)



import inspect
import os
from itertools import dropwhile

import nbformat
import sys
import six

from parade.core.task import ETLTask, SingleSourceETLTask
from parade.command.notebook import NotebookCommand
from parade.command import ParadeCommand


def get_function_body(func):
    source_lines = inspect.getsourcelines(func)[0]
    source_lines = dropwhile(lambda x: x.startswith('@'), source_lines)
    def_line = next(source_lines).strip()
    if def_line.startswith('def ') and def_line.endswith(':'):
        # Handle functions that are not one-liners
        first_line = next(source_lines)
        # Find the indentation of the first line
        indentation = len(first_line) - len(first_line.lstrip())
        return [first_line[indentation:]] + [line[indentation:] for line in source_lines]
    else:
        # Handle single line functions
        return [def_line.rsplit(':')[-1].strip()]


def get_imports(obj):
    source_lines = inspect.getsourcelines(sys.modules[obj.__module__])[0]
    source_lines = dropwhile(lambda x: x.startswith('@'), source_lines)
    import_lines = filter(lambda x: x.startswith('import ') or x.startswith('from '), source_lines)
    return list(import_lines)


class DebugCommand(ParadeCommand):
    requires_workspace = True

    def run_internal(self, context, **kwargs):
        NotebookCommand.init_jupyter(context)

        task = context.get_task(kwargs.get('task'), task_class=ETLTask)
        import_lines = get_imports(task)
        if isinstance(task, SingleSourceETLTask):
            source = task.source
            if isinstance(source, six.string_types):
                body_lines = ['source = """{}"""\n'.format(source)]
            else:
                body_lines = ['source = {}\n'.format({'a': 1})]

            if task.is_source_query:
                body_lines.append('context.load_query(source, conn=\'{}\')\n'.format(task.source_conn))
            else:
                body_lines.append('context.load(source, conn=\'{}\')\n'.format(task.source_conn))

        else:
            body_lines = get_function_body(task.execute_internal)
            body_lines = list(map(lambda x: x.replace('return ', ''), body_lines))
        source_lines = import_lines + ['\n\n'] + body_lines

        notebook = nbformat.v4.new_notebook()
        cell = nbformat.v4.new_code_cell(source=source_lines)
        notebook.cells.append(cell)

        ipynb_path = os.path.join(context.workdir, '.jupyter', 'notebook', task.name + '.ipynb')
        with open(ipynb_path, 'wb') as f:
            ipy_content = nbformat.writes(notebook)
            f.write(ipy_content.encode())
            f.flush()

        os.system('jupyter notebook {}.ipynb'.format(task.name))

    def short_desc(self):
        return 'debug a specified task in notebook'

    def config_parser(self, parser):
        parser.add_argument('task', help='the task to debug')



from ..connection import Connection
import pandas as pd


class FakeConnection(Connection):
    def store(self, df, table, **kwargs):
        print(df)

    def load_query(self, query, **kwargs):
        return pd.DataFrame()

    def load(self, table, **kwargs):
        return pd.DataFrame()




from __future__ import unicode_literals
import decimal
import datetime
import locale
import sys
import dateutil.parser as parser

from .dateparser import DATE_FORMATS, is_date


class CellType(object):
    """ A cell type maintains information about the format
    of the cell, providing methods to check if a type is
    applicable to a given value and to convert a value to the
    type. """

    guessing_weight = 1
    # the type that the result will have
    result_type = None

    def test(self, value):
        """ Test if the value is of the given type. The
        default implementation calls ``cast`` and checks if
        that throws an exception. True or False
        :param value: """
        if isinstance(value, self.result_type):
            return True
        try:
            self.cast(value)
            return True
        except:
            return False

    @classmethod
    def instances(cls):
        return [cls()]

    def cast(self, value):
        """ Convert the value to the type. This may throw
        a quasi-random exception if conversion fails.
        :param value: """
        return value

    def __eq__(self, other):
        return self.__class__ == other.__class__

    def __hash__(self):
        return hash(self.__class__)

    def __repr__(self):
        return self.__class__.__name__.rsplit('Type', 1)[0]


class StringType(CellType):
    """ A string or other unconverted type. """
    result_type = str

    def __init__(self, max_len=0):
        self.max_len = max_len

    def cast(self, value):
        if value is None:
            return None
        if isinstance(value, self.result_type):
            return value
        try:
            return str(value)
        except UnicodeEncodeError:
            return str(value)


class IntegerType(CellType):
    """ An integer field. """
    guessing_weight = 6
    result_type = int

    def __init__(self, length=11):
        self.length = length

    def cast(self, value):
        if value in ('', None):
            return None

        try:
            value = float(value)
        except:
            return locale.atoi(value)

        if value.is_integer():
            return int(value)
        else:
            raise ValueError('Invalid integer: %s' % value)


class DecimalType(CellType):
    """ Decimal number, ``decimal.Decimal`` or float numbers. """
    guessing_weight = 4
    result_type = decimal.Decimal

    def cast(self, value):
        if value in ('', None):
            return None
        try:
            return decimal.Decimal(value)
        except:
            value = locale.atof(value)
            if sys.version_info < (2, 7):
                value = str(value)
            return decimal.Decimal(value)


class FloatType(DecimalType):
    """ FloatType is deprecated """
    pass


class BoolType(CellType):
    """ A boolean field. Matches true/false, yes/no and 0/1 by default,
    but a custom set of values can be optionally provided.
    """
    guessing_weight = 7
    result_type = bool
    true_values = ('yes', 'true')
    false_values = ('no', 'false')

    def __init__(self, true_values=None, false_values=None):
        if true_values is not None:
            self.true_values = true_values
        if false_values is not None:
            self.false_values = false_values

    def cast(self, value):
        s = value.strip().lower()
        if value in ('', None):
            return None
        if s in self.true_values:
            return True
        if s in self.false_values:
            return False
        raise ValueError


class ArrayType(CellType):

    def __init__(self, item_type):
        self.item_type = item_type


class DateType(CellType):
    """ The date type is special in that it also includes a specific
    date format that is used to parse the date, additionally to the
    basic type information. """
    guessing_weight = 3
    formats = DATE_FORMATS
    result_type = datetime.datetime

    def __init__(self, only_date=False, format="%Y-%m-%d %H:%M:%S"):
        self.format = format
        self.only_date = only_date

    @classmethod
    def instances(cls):
        return [cls(v) for v in cls.formats]

    def test(self, value):
        if isinstance(value, str) and not is_date(value):
            return False
        return CellType.test(self, value)

    def cast(self, value):
        if isinstance(value, self.result_type):
            return value
        if value in ('', None):
            return None
        if self.format is None:
            return value
        return datetime.datetime.strptime(value, self.format)

    def __eq__(self, other):
        return (isinstance(other, DateType) and
                self.format == other.format)

    def __repr__(self):
        return "Date(%s)" % self.format

    def __hash__(self):
        return hash(self.__class__) + hash(self.format)


class DateUtilType(CellType):
    """ The date util type uses the dateutil library to
    parse the dates. The advantage of this type over
    DateType is the speed and better date detection. However,
    it does not offer format detection.
    Do not use this together with the DateType"""
    guessing_weight = 3
    result_type = datetime.datetime

    def cast(self, value):
        if value in ('', None):
            return None
        return parser.parse(value)



import json

from sqlalchemy import MetaData, Column, types, Index, Table, select
import datetime
from sqlalchemy.sql import functions, Select

from ..core.task import Task
from ..connection.rdb import RDBConnection


class ParadeRecorder(object):
    _metadata = MetaData()
    _task_table = Table('parade_exec_tasks', _metadata,
                        Column('id', types.BigInteger, autoincrement=True, primary_key=True),
                        Column('project', types.String(128), nullable=False),
                        Column('flow', types.String(128), nullable=False),
                        Column('task', types.String(128), nullable=False),
                        Column('flow_id', types.BigInteger, nullable=False),
                        Column('attributes', types.Text, default="{}"),
                        Column('start_time', types.DateTime, default=datetime.datetime.now),
                        Column('create_time', types.DateTime, default=datetime.datetime.now),
                        Column('commit_time', types.DateTime, default=datetime.datetime.now),
                        Column('update_time', types.DateTime, default=datetime.datetime.now,
                               onupdate=datetime.datetime.now),
                        Column('status', types.Integer, default=Task.STATE_PENDING),
                        Column('message', types.Text, default='OK'),

                        Index('idx_task_create', 'task', 'create_time'),
                        )

    _flow_table = Table('parade_exec_flows', _metadata,
                        Column('id', types.BigInteger, autoincrement=True, primary_key=True),
                        Column('project', types.String(128), nullable=False),
                        Column('flow', types.String(128), nullable=False),
                        Column('tasks', types.Text, nullable=False),
                        Column('create_time', types.DateTime, default=datetime.datetime.now),
                        Column('commit_time', types.DateTime, default=datetime.datetime.now),
                        Column('update_time', types.DateTime, default=datetime.datetime.now,
                               onupdate=datetime.datetime.now),
                        Column('status', types.Integer, default=Task.STATE_EXECUTING),

                        Index('idx_flow_create', 'flow', 'create_time'),
                        )

    def __init__(self, project, conn):
        assert isinstance(conn, RDBConnection)
        self.conn = conn
        self.project = project

    def init_record_tables(self):
        _conn = self.conn.open()
        if not self._task_table.exists(_conn):
            try:
                self._task_table.create(_conn)
            except:
                pass
        if not self._flow_table.exists(_conn):
            try:
                self._flow_table.create(_conn)
            except:
                pass

    def last_success_record(self, task_name):
        _conn = self.conn.open()
        _query = self._task_table.select(). \
            where(self._task_table.c.task == task_name). \
            where(self._task_table.c.status == Task.STATE_SUCCESS). \
            order_by(self._task_table.c.create_time.desc()).limit(1)
        _last_record = _conn.execute(_query).fetchone()

        if _last_record is not None:
            return dict(_last_record)
        return None

    def create_task_record(self, task_name, attributes, flow_id, flow):
        _conn = self.conn.open()

        # 创建待提交checkpoint
        ins = self._task_table.insert().values(project=self.project,
                                               flow_id=flow_id, flow=flow,
                                               task=task_name,
                                               attributes=json.dumps(attributes))
        return _conn.execute(ins).inserted_primary_key[0]

    def create_flow_record(self, flow_name, tasks):
        _conn = self.conn.open()

        # 创建待提交checkpoint
        ins = self._flow_table.insert().values(project=self.project, flow=flow_name, tasks=','.join(tasks))
        return _conn.execute(ins).inserted_primary_key[0]

    def mark_task_start(self, exec_id):
        _conn = self.conn.open()
        sql = self._task_table.update(). \
            where(self._task_table.c.id == exec_id). \
            values(status=Task.STATE_EXECUTING, start_time=functions.now(), update_time=functions.now())
        _conn.execute(sql)

    def mark_task_success(self, exec_id, msg='OK'):
        _conn = self.conn.open()
        sql = self._task_table.update(). \
            where(self._task_table.c.id == exec_id). \
            values(status=Task.STATE_SUCCESS, message=str(msg), commit_time=functions.now(), update_time=functions.now())
        _conn.execute(sql)

    def mark_task_failed(self, exec_id, err):
        _conn = self.conn.open()
        sql = self._task_table.update(). \
            where(self._task_table.c.id == exec_id). \
            values(status=Task.STATE_FAILED, message=str(err), update_time=functions.now())
        _conn.execute(sql)

    def mark_task_cancelled(self, exec_id, failed_deps):
        _conn = self.conn.open()
        sql = self._task_table.update(). \
            where(self._task_table.c.id == exec_id). \
            values(status=Task.STATE_CANCELLED, message='canceled for failed dependencies [{}]'.format(failed_deps), update_time=functions.now())
        _conn.execute(sql)

    def mark_flow_success(self, flow_id):
        _conn = self.conn.open()
        sql = self._flow_table.update(). \
            where(self._flow_table.c.id == flow_id). \
            values(status=Task.STATE_SUCCESS, commit_time=functions.now(), update_time=functions.now())
        _conn.execute(sql)

    def mark_flow_failed(self, flow_id):
        _conn = self.conn.open()
        sql = self._flow_table.update(). \
            where(self._flow_table.c.id == flow_id). \
            values(status=Task.STATE_FAILED, update_time=functions.now())
        _conn.execute(sql)

    def load_flow_by_id(self, exec_id):
        _conn = self.conn.open()
        _query = self._flow_table.select(). \
            where(self._flow_table.c.id == exec_id)
        exec_flow = _conn.execute(_query).fetchone()

        if exec_flow is not None:
            raw = dict(exec_flow)
            raw['create_time'] = str(raw['create_time'])
            raw['update_time'] = str(raw['update_time'])
            raw['commit_time'] = str(raw['commit_time'])
            return raw
        return None

    def load_flows(self, flow=None, executing=None, page_size=0, page_no=1):
        from sqlalchemy import func
        query = self._flow_table.select()
        if flow is not None:
            query = query.where(self._flow_table.c.flow == flow)
        if executing is not None:
            query = query.where(
                self._flow_table.c.status == 0) if executing else query.where(
                self._flow_table.c.status > 0)

        total = None
        if page_size > 0:
            count_query = query.with_only_columns([func.count(self._flow_table.c.id)])
            query = query.limit(page_size).offset((page_no - 1) * page_size)
            total = self.conn.open().execute(count_query).scalar()

        query = query.order_by(self._flow_table.c.create_time.desc())
        df = self.conn.load_query(str(query.compile(compile_kwargs={"literal_binds": True})))

        df['create_time'] = df['create_time'].apply(lambda ts: str(ts))
        df['commit_time'] = df['commit_time'].apply(lambda ts: str(ts))
        df['update_time'] = df['update_time'].apply(lambda ts: str(ts))

        if not total:
            total = len(df)
        return {'data': json.loads(df.to_json(orient='records')), 'total': total}

    def load_flow_tasks(self, flow_id, page_size=0, page_no=1):
        query = self._task_table.select().where(self._task_table.c.flow_id == flow_id)
        if page_size > 0:
            query = query.limit(page_size).offset((page_no - 1) * page_size)
        df = self.conn.load_query(str(query.compile(compile_kwargs={"literal_binds": True})))

        df['start_time'] = df['start_time'].apply(lambda ts: str(ts))
        df['create_time'] = df['create_time'].apply(lambda ts: str(ts))
        df['commit_time'] = df['commit_time'].apply(lambda ts: str(ts))
        df['update_time'] = df['update_time'].apply(lambda ts: str(ts))

        tasks = json.loads(df.to_json(orient='records'))
        return tasks



import urllib
from os.path import join

from .. import ParadeCommand


class InstallCommand(ParadeCommand):
    """
    The install command will install acontrib component into current workspace
    """

    CONTRIB_REPO_PATH = 'https://raw.githubusercontent.com/bailaohe/parade-contrib/master/{}/{}.py'

    requires_workspace = True

    def short_desc(self):
        return 'install a contrib component into current workspace'

    def config_parser(self, parser):
        parser.add_argument('component_type',
                            help='the type of the component to install, "connection" or "flowstore" at this moment')
        parser.add_argument('component_key', help='the key of the component to install')

    def run_internal(self, context, **kwargs):
        component_type = kwargs['component_type']
        component_key = kwargs['component_key']

        install_path = join(context.workdir, context.name, "contrib", component_type)
        urllib.request.urlretrieve(InstallCommand.CONTRIB_REPO_PATH.format(component_type, component_key),
                                   join(install_path, component_key + '.py'))




import os
from . import ParadeCommand


def _init_web():
    from flask import Blueprint
    web = Blueprint('web', __name__)

    @web.route("/")
    def route():
        from flask import render_template
        return render_template("index.html")

    return web


def _create_app(context):
    from flask import Flask
    from flask_cors import CORS
    from flask_socketio import SocketIO

    template_dir = os.path.join(context.workdir, 'web')
    static_dir = os.path.join(context.workdir, 'web', 'static')

    app = Flask(context.name, template_folder=template_dir, static_folder=static_dir)
    CORS(app)

    app.parade_context = context

    from ..api import parade_blueprint
    app.register_blueprint(parade_blueprint)

    web_blueprint = _init_web()
    app.register_blueprint(web_blueprint)

    socketio = SocketIO(app, async_mode='threading')
    sio = socketio.server

    @sio.on('connect', namespace='/exec')
    def connect(sid, environ):
        pass

    @sio.on('query', namespace='/exec')
    def query(sid, data):
        exec_id = data
        sio.enter_room(sid, str(exec_id), namespace='/exec')
        sio.emit('reply', exec_id, namespace='/exec')

    context.webapp = app

    return app, socketio


class ServerCommand(ParadeCommand):
    requires_workspace = True

    def run_internal(self, context, **kwargs):
        port = int(kwargs.get('port', 5000))
        app, socketio = _create_app(context)
        debug = context.conf.get_or_else('debug', False)

        socketio.run(app, host="0.0.0.0", port=port, debug=debug, log_output=False)

    def short_desc(self):
        return 'start a parade api server'

    def config_parser(self, parser):
        parser.add_argument('-p', '--port', default=5000, help='the port of parade server')



from os.path import join
from shutil import copy

import parade
from parade.command import ParadeCommand
from parade.utils.template import string_camelcase, render_templatefile


class GenTaskCommand(ParadeCommand):
    """
    The gentask command will create a task of specified type in current workspace
    """
    requires_workspace = True

    @property
    def aliases(self):
        return ['new']

    def short_desc(self):
        return 'generate a task skeleton with specified type'

    def config_parser(self, parser):
        parser.add_argument('task', nargs='*', help='the name of the task to generate')
        parser.add_argument('-t', '--task_type', dest='task_type', help='the type of the task to generate', required=True)

    def run_internal(self, context, **kwargs):
        task_names = kwargs['task']
        task_type = kwargs['task_type']

        target_path = join(context.workdir, context.name, "task")

        source_tpl = task_type + ".py.tpl"

        for task_name in task_names:
            target_tpl = task_name + ".py"

            source_tplfile = join(self.templates_dir, source_tpl)
            target_tplfile = join(target_path, target_tpl)
            copy(source_tplfile, target_tplfile)

            render_templatefile(target_tplfile, TaskName=string_camelcase(task_name))

    @property
    def templates_dir(self):
        _templates_base_dir = join(parade.__path__[0], 'template')
        return join(_templates_base_dir, 'task')




import json

from flask import request, make_response
from flask_restful import Api, reqparse

from . import parade_blueprint, ParadeResource, catch_parade_error
from ..connection.localfile import LocalFile
from ..core.task import ETLTask

api = Api(parade_blueprint, catch_all_404s=True)
parser = reqparse.RequestParser()


class DataAPI(ParadeResource):
    """
    The api blue print to execute etl task
    """

    @catch_parade_error
    def get(self, name):
        # data in string format and you have to parse into dictionary
        data_task = self.context.get_task(name, task_class=ETLTask)
        return data_task.info

    @catch_parade_error
    def post(self, name):
        # data in string format and you have to parse into dictionary
        complete = request.args.get('complete', type=bool, default=False)
        export = request.args.get('export', default=None)

        task_args = request.get_json() or {}
        data_task = self.context.get_task(name, task_class=ETLTask)

        df = data_task.execute_internal(self.context, **task_args)

        if export:
            export_io, export_file = LocalFile.export(df, name, export_type=export)
            response = make_response(export_io.getvalue())
            response.headers["Content-Disposition"] = "attachment; filename=" + str(export_file)
            return response

        data = json.loads(df.to_json(orient='records'))
        return dict({'data': data}, **data_task.attributes) if complete else data


api.add_resource(DataAPI, '/api/data/<name>')



from . import Notifier
from ..utils.log import parade_logger as logger


class ParadeNotifier(Notifier):
    """
    This is *JUST A PLACEHOLDER* implementation
    """
    def notify_error(self, task, reason, **kwargs):
        logger.error('task {task} failed: {reason}'.format(task=task, reason=reason))

    def notify_success(self, task, **kwargs):
        logger.error('task {task} succeeded'.format(task=task))



# -*- coding:utf-8 -*-
import sys
import traceback

from flask import Blueprint, current_app
from flask_restful import Resource, abort

from ..error import ParadeError


class ParadeResource(Resource):
    """
    The api blue print to execute etl task
    """

    def __init__(self):
        self.context = current_app.parade_context


def catch_parade_error(func):
    def wrapper(*args, **kw):
        try:
            return func(*args, **kw)
        except ParadeError as e:
            if current_app.debug:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                stack_info = traceback.format_exception(exc_type, exc_value, exc_traceback)
                abort(e.status, code=e.code, message=e.reason, traceback=stack_info)
            else:
                abort(e.status, code=e.code, message=e.reason)
        except Exception as e:
            if current_app.debug:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                stack_info = traceback.format_exception(exc_type, exc_value, exc_traceback)
                abort(500, code=0, message=str(e), traceback=stack_info)
            else:
                abort(500, code=0, message=str(e))
    return wrapper

parade_blueprint = Blueprint('parade', __name__)

from . import task
from . import flow
from . import exec
from . import data



from parade.command import ParadeCommand


class ListTaskCommand(ParadeCommand):

    requires_workspace = True

    @property
    def aliases(self):
        return ['ls']

    def run_internal(self, context, **kwargs):
        tasks = context.list_tasks()
        print('Workspace [{}] has {} task(s):'.format(context.name, len(tasks)))
        for task_id, task in enumerate(tasks):
            print("#{}: {}".format(task_id + 1, task))

    def short_desc(self):
        return 'list the tasks created'

    def config_parser(self, parser):
        pass




from parade.command import ParadeCommand
from parade.core.engine import Engine


class ExecCommand(ParadeCommand):
    """
    The exec command to run a flow or a set tasks,
    if the tasks to execute have dependencies on each other,
    parade will handle them correctly
    """
    requires_workspace = True

    @property
    def aliases(self):
        return ['execute', 'exec']

    def run_internal(self, context, **kwargs):
        engine = Engine(context)

        flow = kwargs.get('flow')
        force = kwargs.get('force')

        return engine.execute_async(flow_name=flow, force=force)

    def short_desc(self):
        return 'execute a set of tasks'

    def config_parser(self, parser):
        parser.add_argument('--force', action="store_true", help='force the tasks to execute')
        parser.add_argument('flow', help='the flow to schedule')



FLOW_NOT_FOUND = (100, 404, 'flow [{flow}] not found')

class ParadeError(RuntimeError):
    """
    base exception of parade
    """

    code = 0
    status = 200
    message = 'OK'
    args = {}

    def __init__(self, **err_args):
        self.args = err_args

    @property
    def reason(self):
        return self.message.format(**self.args)



import asyncio
from asyncio import queues

from ..core.task import Flow, Task
from ..flowrunner import FlowRunner
from ..utils.log import logger


class ParadeFlowRunner(FlowRunner):
    # the thread pool to convert block execution of task into async process
    wait_queue = None
    exec_queue = None
    executing_flow = None
    executing_flow_id = 0
    kwargs = {}

    def initialize(self, context, conf):
        FlowRunner.initialize(self, context, conf)

    def submit(self, flow, flow_id=0, **kwargs):
        """
        execute a set of tasks with DAG-topology into consideration
        :param task_names: the tasks to form DAG
        :return:
        """

        assert isinstance(flow, Flow)
        self.executing_flow = flow
        self.executing_flow_id = flow_id
        self.kwargs = kwargs

        for task_name in self.executing_flow.tasks:
            self.context.get_task(task_name).pending(self.context, flow_id, flow.name)
        io_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(io_loop)
        self.wait_queue = queues.Queue()
        self.exec_queue = queues.Queue()
        ret = io_loop.run_until_complete(self.execute_dag_ioloop())
        io_loop.close()
        return ret

    @asyncio.coroutine
    def daemon_loop(self):
        yield from self.execute_dag_ioloop()

    @asyncio.coroutine
    def execute_dag_ioloop(self):
        """
        the async process to execute task DAG
        :return:
        """

        # add to wait queue, waiting to execute
        yield from self.wait_queue.put(set(self.executing_flow.tasks))

        executing, successed, failed = set(), set(), set()

        @asyncio.coroutine
        def _produce_tasks():
            """
            the inner async procedure of task producer
            :return:
            """
            # wait until all tasks in executing queue are done
            yield from self.exec_queue.join()

            # reset the *executing* and *done* set
            executing.clear()
            successed.clear()
            failed.clear()

            # retrieve the task-DAG from wait-queue to exec-queue
            sched_task_names = yield from self.wait_queue.get()
            for sched_task_name in sched_task_names:
                yield from self.exec_queue.put(sched_task_name)
            self.wait_queue.task_done()

        @asyncio.coroutine
        def _consume_task():
            """
            the inner async procedure of task consumers
            :return:
            """
            next_task_name = self.exec_queue.get_nowait()

            logger(exec=self.executing_flow_id, flow=self.executing_flow.name).info("pick up task [{}] ...".format(next_task_name))
            try:
                if next_task_name in executing:
                    logger(exec=self.executing_flow_id, flow=self.executing_flow.name).info("task [{}] is executing, pass ...".format(next_task_name))
                    return

                next_task = self.context.get_task(next_task_name)
                task_deps = self.executing_flow.deps.get(next_task_name, set())

                done_deps = set(filter(lambda x: x in successed, task_deps))
                fail_deps = set(filter(lambda x: x in failed, task_deps))

                if len(task_deps) == len(done_deps):
                    # all dependencies are done
                    # submit the task to threading pool to execute
                    if len(task_deps) > 0:
                        logger(exec=self.executing_flow_id, flow=self.executing_flow.name).info("all dependant task(s) of task {} is done".format(next_task_name))
                    executing.add(next_task_name)

                    logger(exec=self.executing_flow_id, flow=self.executing_flow.name).info("task [{}] start executing ...".format(next_task_name))
                    next_task.execute(self.context, flow_id=self.executing_flow_id, flow=self.executing_flow,
                                      **self.kwargs)
                    if next_task.result_code == Task.RET_CODE_SUCCESS:
                        logger(exec=self.executing_flow_id, flow=self.executing_flow.name).info("task [{}] executed successfully".format(next_task_name))
                        successed.add(next_task_name)
                    else:
                        logger(exec=self.executing_flow_id, flow=self.executing_flow.name).info("task [{}] executed failed".format(next_task_name))
                        failed.add(next_task_name)

                elif len(fail_deps) > 0:
                    logger(exec=self.executing_flow_id, flow=self.executing_flow.name).info("task [{}] canceled since its dependencies [{}] failed".format(next_task_name, fail_deps))
                    executing.add(next_task_name)
                    failed.add(next_task_name)
                    next_task.cancel(self.context, fail_deps)
                else:
                    # otherwise, re-put the task into the end of the queue
                    # sleep for 1 second
                    yield from self.exec_queue.put(next_task_name)
                    yield from asyncio.sleep(1)
            except Exception as e:
                logger(exec=self.executing_flow_id, flow=self.executing_flow.name).exception(str(e))
            finally:
                self.exec_queue.task_done()

        @asyncio.coroutine
        def consumer():
            while True:
                yield from _consume_task()

        @asyncio.coroutine
        def producer():
            yield from _produce_tasks()

        # we use a single producer within the main-thread
        yield from producer()

        try:
            yield from consumer()
        except:
            pass

        yield from self.exec_queue.join()
        assert len(executing) == len(successed) + len(failed)

        retcode = Task.RET_CODE_SUCCESS if len(failed) == 0 else Task.RET_CODE_FAIL

        if retcode == Task.RET_CODE_SUCCESS:
            self.context.on_flow_success(self.executing_flow_id)
        else:
            self.context.on_flow_failed(self.executing_flow_id)

        return retcode



import re
import string
from importlib import import_module
from os.path import join, exists, abspath
from shutil import move, ignore_patterns

import parade
from ..utils.misc import copytree
from . import ParadeCommand
from ..utils.template import string_camelcase, render_templatefile

TEMPLATES_TO_RENDER = (
    ('parade.bootstrap.yml',),
    ('parade.config.yml',),
)

IGNORE = ignore_patterns('*.pyc', '.svn', '.git')


class InitCommand(ParadeCommand):
    """
    The init command will create a workspace with template.
    The workspace will hold all user's work (tasks and flows)
    and customized configurations.
    """
    requires_workspace = False

    def short_desc(self):
        return 'init a workspace to work with'

    def config_parser(self, parser):
        parser.add_argument('workspace', help='the workspace to initialize')
        parser.add_argument('workdir', nargs='?', help='the directory of the workspace')

    @staticmethod
    def _is_valid_name(workspace_name):
        """
        check if the workspace name is valid
        :param workspace_name: the specified workspace name
        :return:
        """

        def _module_exists(module_name):
            try:
                import_module(module_name)
                return True
            except ImportError:
                return False

        if not re.search(r'^[_a-zA-Z]\w*$', workspace_name):
            print('Error: Workspace names must begin with a letter and contain' \
                  ' only\nletters, numbers and underscores')
        elif _module_exists(workspace_name):
            print('Error: Module %r already exists' % workspace_name)
        else:
            return True
        return False

    def run_internal(self, context, **kwargs):
        project_name = kwargs['workspace']
        project_dir = project_name

        if 'workdir' in kwargs and kwargs['workdir']:
            project_dir = kwargs['workdir']

        if exists(join(project_dir, 'parade.parade.bootstrap.yml')):
            self.exitcode = 1
            print('Error: parade.parade.bootstrap.yml already exists in %s' % abspath(project_dir))
            return

        if not self._is_valid_name(project_name):
            self.exitcode = 1
            return

        copytree(self.templates_dir, abspath(project_dir), IGNORE)
        move(join(project_dir, 'module'), join(project_dir, project_name))
        for paths in TEMPLATES_TO_RENDER:
            path = join(*paths)
            tplfile = join(project_dir,
                           string.Template(path).substitute(project_name=project_name))
            render_templatefile(tplfile, project_name=project_name, project_path=abspath(project_dir),
                                ProjectName=string_camelcase(project_name))
        move(join(project_dir, 'parade.config.yml'), join(project_dir, project_name + '-default-1.0.yml'))
        print("New Parade workspace %r, using template directory %r, created in:" % \
              (project_name, self.templates_dir))
        print("    %s\n" % abspath(project_dir))
        print("You can start your first task with:")
        print("    cd %s" % project_dir)
        print("    parade gentask your_task -t etl")

    @property
    def templates_dir(self):
        _templates_base_dir = join(parade.__path__[0], 'template')
        return join(_templates_base_dir, 'workspace')



from github import Github

from .. import ParadeCommand


class SearchCommand(ParadeCommand):
    """
    The search command will query the prarade-contrib on github for a contributed component
    """
    requires_workspace = False

    def short_desc(self):
        return 'search a contrib component'

    def config_parser(self, parser):
        parser.add_argument('query', help='the query to lookup the contrib component')

    def run_internal(self, context, **kwargs):
        query = kwargs['query']

        g = Github('parade-user', 'parade123')
        hits = g.search_code(query, repo="bailaohe/parade-contrib", language='python', **{'in': 'path'})

        if hits.totalCount <= 0:
            print("Not found!")
        else:
            print("Found {} components:".format(hits.totalCount))
            for hit in hits:
                component_path = hit.path[0:hit.path.find('.py')]
                print(component_path)



from __future__ import unicode_literals
from collections import defaultdict

from .stdtypes import StringType, IntegerType, DecimalType, BoolType, DateType

_STRING_TYPE = StringType()
_INTEGER_TYPE = IntegerType()
_DECIMAL_TYPE = DecimalType()
_BOOLEAN_TYPE = BoolType()
_DATE_TYPE = DateType()


def str_to_stdtype(expr):
    norm_expr = expr.lower()
    if norm_expr.startswith('integer'):
        import re
        match_result = re.match(r'integer\((\d+)\)', norm_expr)
        if match_result is not None:
            return IntegerType(length=int(match_result.group(1)))
        return _INTEGER_TYPE
    if norm_expr == 'decimal':
        return _DECIMAL_TYPE
    if norm_expr == 'date':
        return _DATE_TYPE
    if norm_expr == 'bool' or norm_expr == 'boolean':
        return _BOOLEAN_TYPE
    if norm_expr.startswith('string'):
        match_result = re.match(r'string\((\d+)\)', norm_expr)
        if match_result is not None:
            return StringType(max_len=int(match_result.group(1)))
        return _STRING_TYPE
    raise RuntimeError("Unsupported data type [" + expr + "]")


def str_to_sqltype(expr):
    import re
    import sqlalchemy.types as sqltypes
    norm_expr = expr.lower()
    if norm_expr.startswith('integer'):
        match_result = re.match(r'integer\((\d+)\)', norm_expr)
        if match_result is not None:
            return sqltypes.BIGINT() if int(match_result.group(1)) > 11 else sqltypes.INTEGER()
        return sqltypes.BIGINT()
    if norm_expr == 'decimal':
        return sqltypes.DECIMAL()
    if norm_expr == 'date':
        return sqltypes.DATETIME()
    if norm_expr == 'bool' or norm_expr == 'boolean':
        return sqltypes.BOOLEAN()
    if norm_expr.startswith('string'):
        match_result = re.match(r'string\((\d+)\)', norm_expr)
        if match_result is not None:
            maxlen = int(match_result.group(1))
            return sqltypes.VARCHAR(maxlen) if maxlen < 65536 else sqltypes.TEXT
        return sqltypes.TEXT()
    raise RuntimeError("Unsupported data type [" + expr + "]")


def sqltype_to_stdtype(sqltype):
    import sqlalchemy.types as sqltypes
    if isinstance(sqltype, (sqltypes.VARCHAR, sqltypes.CHAR, sqltypes.TEXT, sqltypes.Enum, sqltypes.String)):
        return _STRING_TYPE
    if isinstance(sqltype, (sqltypes.DATETIME, sqltypes.DATE, sqltypes.TIME, sqltypes.TIMESTAMP)):
        return _DATE_TYPE
    if isinstance(sqltype, (sqltypes.INTEGER, sqltypes.BIGINT, sqltypes.SMALLINT, sqltypes.Integer)):
        return _INTEGER_TYPE
    if isinstance(sqltype, (sqltypes.REAL, sqltypes.DECIMAL, sqltypes.NUMERIC, sqltypes.FLOAT)):
        return _DECIMAL_TYPE
    if isinstance(sqltype, sqltypes.BOOLEAN):
        return _BOOLEAN_TYPE


def stdtype_to_sqltype(stdtype):
    import sqlalchemy.types as sqltypes
    if isinstance(stdtype, stdtypes.StringType):
        return sqltypes.VARCHAR(length=stdtype.max_len) if 0 < stdtype.max_len < 65536 else sqltypes.TEXT()
    if isinstance(stdtype, stdtypes.BoolType):
        return sqltypes.BOOLEAN()
    if isinstance(stdtype, stdtypes.DateType):
        return sqltypes.DATE() if stdtype.only_date else sqltypes.TIMESTAMP()
    if isinstance(stdtype, stdtypes.IntegerType):
        return sqltypes.BIGINT() if stdtype.length > 11 else sqltypes.INTEGER()
    if isinstance(stdtype, stdtypes.DecimalType):
        return sqltypes.DECIMAL()
    if isinstance(stdtype, stdtypes.ArrayType):
        return sqltypes.ARRAY(item_type=stdtype.item_type)


def type_guess(df, strict=False):
    """ The type guesser aggregates the number of successful
    conversions of each column to each type, weights them by a
    fixed type priority and select the most probable type for
    each column based on that figure. It returns a list of
    ``CellType``. Empty cells are ignored.
    Strict means that a type will not be guessed
    if parsing fails for a single cell in the column.
    :param df:
    :param strict: """
    guesses = []
    types = [_STRING_TYPE, _INTEGER_TYPE, _DECIMAL_TYPE, _BOOLEAN_TYPE, _DATE_TYPE]
    type_instances = [i for t in types for i in t.instances()]
    if strict:
        at_least_one_value = []
        for row in df.iterrows():
            # ri = row[0]
            cells = row[1]
            diff = len(cells) - len(guesses)
            for _ in range(diff):
                type_dict = {}
                for stdtype in type_instances:
                    type_dict[stdtype] = 0
                guesses.append(type_dict)
                at_least_one_value.append(False)
            for ci, cell in enumerate(cells.iteritems()):
                # key = cell[0]
                val = str(cell[1])
                if not val:
                    continue
                at_least_one_value[ci] = True
                for stdtype in list(guesses[ci].keys()):
                    if not stdtype.test(val):
                        guesses[ci].pop(stdtype)
        # no need to set guessing weights before this
        # because we only accept a type if it never fails
        for i, guess in enumerate(guesses):
            for stdtype in guess:
                guesses[i][stdtype] = stdtype.guessing_weight
        # in case there were no values at all in the column,
        # we just set the guessed type to string
        for i, v in enumerate(at_least_one_value):
            if not v:
                guesses[i] = {StringType(): 0}
    else:
        for row in df.iterrows():
            # i = row[0]
            cells = row[1]
            diff = len(cells) - len(guesses)
            for _ in range(diff):
                guesses.append(defaultdict(int))
            for i, cell in enumerate(cells.iteritems()):
                # add string guess so that we have at least one guess
                guesses[i][StringType()] = guesses[i].get(StringType(), 0)
                # key = cell[0]
                val = str(cell[1])
                if not val:
                    continue
                for stdtype in type_instances:
                    if stdtype.test(val):
                        guesses[i][stdtype] += stdtype.guessing_weight
    _columns = []
    for guess in guesses:
        # this first creates an array of tuples because we want the types to be
        # sorted. Even though it is not specified, python chooses the first
        # element in case of a tie
        # See: http://stackoverflow.com/a/6783101/214950
        guesses_tuples = [(t, guess[t]) for t in type_instances if t in guess]
        _columns.append(max(guesses_tuples, key=lambda t_n: t_n[1])[0])
    return dict(zip(df.columns, _columns))



import pandas as pd
from ..core import Plugin
from sqlalchemy import create_engine, MetaData, Column, Table
from sqlalchemy.exc import NoSuchTableError

from parade.utils.log import parade_logger as logger
from . import Connection


class RDBConnection(Connection):
    def __init__(self):
        Plugin.__init__(self)

    def open(self):
        uri = self.datasource.uri
        if uri is None:
            authen = None
            uripart = self.datasource.host + ':' + str(self.datasource.port) + '/' + self.datasource.db
            if self.datasource.user is not None:
                authen = self.datasource.user
            if authen is not None and self.datasource.password is not None:
                authen += ':' + self.datasource.password
            if authen is not None:
                uripart = authen + '@' + uripart
            uri = self.datasource.protocol + '://' + uripart + '?charset=utf8'
        return create_engine(uri, encoding="utf-8", pool_size=16, pool_recycle=300)

    def load(self, table, **kwargs):
        return self.load_query('select * from {}'.format(table))

    def load_query(self, query, **kwargs):
        conn = self.open()
        df = pd.read_sql_query(query, con=conn)

        # 优化内存使用
        # 1.使用子类型优化数字列
        df = df.apply(pd.to_numeric, errors='ignore', downcast='unsigned')
        # 2.使用分类来优化对象类型
        df_obj = df.select_dtypes(include=['object'])
        for col in df_obj.columns:
            num_unique_values = len(df_obj[col].unique())
            num_total_values = len(df_obj[col])
            if num_unique_values / num_total_values < 0.5:
                df[col] = df[col].astype('category')

    def store(self, df, table, **kwargs):
        assert isinstance(df, pd.DataFrame), "Invalid data type"
        if_exists = kwargs.get('if_exists', 'fail')
        chunksize = kwargs.get('chunksize', 10000)
        pkey = kwargs.get('pkey', None)
        indexes = kwargs.get('indexes', [])
        checkpoint_column = kwargs.get('checkpoint_column', None)
        checkpoint = kwargs.get('checkpoint')
        last_checkpoint = kwargs.get('last_checkpoint')

        _conn = self.open()

        try:
            if if_exists == 'append' or if_exists == 'update':
                target_table = Table(table, MetaData(), autoload=True, autoload_with=_conn)
                assert checkpoint_column is not None, "checkpoint_column is required in update mode!"
                assert (isinstance(checkpoint_column, tuple) and len(checkpoint_column) == 2) or isinstance(
                    checkpoint_column, str), "checkpoint_column can only be str or 2-tuple!"

                if isinstance(checkpoint_column, tuple):
                    (create_time_column, update_time_column) = checkpoint_column
                else:
                    create_time_column = checkpoint_column
                    update_time_column = checkpoint_column

                # delete extra records over last checkpoint in append/update mode
                clear_ins = target_table.delete().where(Column(update_time_column) >= last_checkpoint)
                _conn.execute(clear_ins)

                if if_exists == 'update':
                    assert pkey is not None, "primary key is required in update mode!"
                    assert isinstance(pkey, str), "update mode only support single primary key"
                    update_df = df[df[create_time_column] < last_checkpoint]
                    if not update_df.empty:
                        logger.info(table + ": find {} records to update".format(len(update_df)))
                        update_keys = list(update_df[pkey])
                        delete_ins = target_table.delete().where(Column(pkey).in_(update_keys))
                        _conn.execute(delete_ins)
                    if_exists = 'append'
        except NoSuchTableError:
            if_exists = 'replace'

        schema = None
        if table.find('.') >= 0:
            toks = table.split('.', 1)
            schema = toks[0]
            table = toks[1]

        float_columns = list(df.select_dtypes(include=['float64', 'float']).keys())
        if len(float_columns) > 0:
            logger.warn(table +
                        ": Detect columns with float types {}, you better check if this is caused by NAN-integer "
                        "column issue of pandas!".format(
                            list(float_columns)))

        typehints = dict()
        obj_columns = list(df.select_dtypes(include=['object']).keys())

        if len(obj_columns) > 0:
            logger.warn(table +
                        ": Detect columns with object types {}, which is automatically converted to *VARCHAR(256)*, "
                        "you can override this by specifying type hints!".format(
                            list(obj_columns)))
        import sqlalchemy.types as sqltypes
        typehints.update(dict((k, sqltypes.VARCHAR(256)) for k in obj_columns))

        # TODO: upddate typehints with user-specified one
        _typehints = kwargs.get('typehints', {})
        from parade.type import stdtype_to_sqltype
        for col, stdtype in _typehints.items():
            logger.info(table + ": Column [{}] is set to type [{}]".format(col, str(stdtype)))
            typehints[col] = stdtype_to_sqltype(stdtype)

        def _chunks(_df, _chunksize):
            """Yield successive n-sized chunks from l."""
            for i in range(0, len(_df), _chunksize):
                yield df[i:i + _chunksize]

        # still write to database for empty dataframe
        if df.empty:
            df.to_sql(name=table, con=_conn, index=False, schema=schema, if_exists=if_exists, dtype=typehints)
            logger.warn(table + ": Write to {}: empty dataframe".format(table))
        else:
            for idx, chunk in enumerate(_chunks(df, chunksize)):
                if_exists_ = 'append' if idx > 0 else if_exists
                chunk.to_sql(name=table, con=_conn, index=False, schema=schema, if_exists=if_exists_, dtype=typehints)
                logger.info(table + ": Write to {}: rows #{}-#{}".format(table, idx * chunksize, (idx + 1) * chunksize))

        if if_exists == 'replace':
            if pkey:
                pkeys = pkey if isinstance(pkey, str) else ','.join(pkey)
                _conn.execute('ALTER TABLE {} ADD PRIMARY KEY ({})'.format(table, pkeys))

            for index in indexes:
                index_str = index if isinstance(index, str) else ','.join(index)
                index_name = index if isinstance(index, str) else '_'.join(index)
                _conn.execute('ALTER TABLE {} ADD INDEX idx_{} ({})'.format(table, index_name, index_str))



from ..core import Plugin


class FlowStore(Plugin):
    """
    The flowstore to support task-flow scheduling
    """
    def create(self, flow, *tasks, deps=None, **kwargs):
        """
        schedule the task-flow
        :param dag_key: the flow name
        :param cron: the cron string to schedule the flow
        :return:
        """
        raise NotImplementedError

    def list(self):
        raise NotImplementedError

    def load(self, flow):
        raise NotImplementedError

    def delete(self, flow_name):
        """
        unschedule the task-flow
        :param flow_name: the flow name
        :return:
        """
        pass



import json
import sys
import time

import pandas as pd

from ..connection import Connection
from ..utils.log import logger
from ..utils.timeutils import datetime_str_to_timestamp, timestamp_to_datetime


class Task(object):
    """
    The task object executed by the parade engine
    """
    RET_CODE_SUCCESS = 0
    RET_CODE_FAIL = 1

    STATE_INIT = (0, 'Created')[0]
    STATE_PENDING = (1, 'Pending')[0]
    STATE_EXECUTING = (2, 'Executing')[0]
    STATE_SUCCESS = (3, 'Succeeded')[0]
    STATE_FAILED = (4, 'Failed')[0]
    STATE_CANCELLED = (5, 'Cancelled')[0]

    def __init__(self):
        """
        _result: the cached task result after execution
        _attributes: the attribute of the task
        :return:
        """
        self._result = None
        self._attributes = {}
        self._result_code = self.RET_CODE_SUCCESS
        self.exec_id = 0
        self._state = self.STATE_INIT
        self.flow_id = 0
        self._flow = None

    @property
    def name(self):
        """
        get the identifier of the task, the default is the class name of task
        :return: the task identifier
        """
        # class_name = self.__class__.__name__
        # task_name = ''
        # for _s_ in class_name:
        #     task_name += _s_ if _s_.islower() else '_' + _s_.lower()
        # return task_name[task_name.startswith("_") and 1:]
        return self.__module__.split('.')[-1]

    @property
    def deps(self):
        """
        a string-array to specified the dependant tasks has to be completed before this one
        :return:
        """
        return set()

    @property
    def attributes(self):
        """
        the attributes to transferred to the task result
        :return:
        """
        return self._attributes

    @property
    def result(self):
        """
        the task result
        :return:
        """
        return self._result

    @property
    def result_code(self):
        """
        the task result code
        :return:
        """
        return self._result_code

    def set_attribute(self, key, val):
        """
        set the task attribute
        :param key:
        :param val:
        :return:
        """
        self._attributes[key] = val

    @property
    def notify_success(self):
        return False

    @property
    def notify_fail(self):
        return True

    def on_start(self, context, **kwargs):
        """
        start a checkpoint transaction before executing the task
        :param context:
        :return:
        """

        assert self._state == self.STATE_PENDING
        return self.STATE_EXECUTING

    def pending(self, context, flow_id=0, flow=None):
        """
        Initialize the task to *pending* state
        :return:
        """
        self.flow_id = flow_id
        self._flow = flow

        self._state = self.STATE_PENDING
        self.on_pending(context)
        self.exec_id = context.on_task_pending(self.name, self.attributes, flow_id, flow)

    def on_pending(self, context):
        pass

    def cancel(self, context, fail_deps):
        self._state = self.STATE_CANCELLED
        self.on_cancel(context)
        context.on_task_cancelled(self, fail_deps)

    def on_cancel(self, context):
        pass

    def execute(self, context, **kwargs):
        """
        the execution process of the etl task
        :param context:
        :param kwargs:
        :return:
        """

        new_state = self.on_start(context, **kwargs)
        self._state = new_state

        try:
            # run if
            # 1. task record is inited. or
            # 2. checkpoint connection is not specified
            if self._state == self.STATE_EXECUTING:
                context.on_task_start(self)
                self._result = self.execute_internal(context, **kwargs)
                self._state = self.STATE_SUCCESS
                self.on_commit(context)

            assert self._state == self.STATE_SUCCESS
            self._result_code = self.RET_CODE_SUCCESS
            context.on_task_success(self)

        except Exception as e:
            logger(exec=self.flow_id, flow=self._flow, task=self.name).exception(str(e))
            self._state = self.STATE_FAILED
            self._result_code = self.RET_CODE_FAIL
            self._result = e
            self.on_rollback(context)
            context.on_task_failed(self, e)

    def on_commit(self, context, **kwargs):
        pass

    def on_rollback(self, context, **kwargs):
        pass

    def execute_internal(self, context, **kwargs):
        """
        the execution process of the task
        :param context: the executor context
        :param kwargs: the task arguments
        :return: the task result
        """
        raise NotImplementedError

    @property
    def info(self):
        return {
            'name': self.name,
            'class': type(self).__name__,
            'module': type(self).__module__,
            'bases': [x.__module__ + '.' + x.__name__ for x in type(self).__bases__],
            # 'attrs': list(filter(lambda x: x not in dir(Task) and not x.startswith('_'), dir(task)))
            'attrs': list(filter(lambda x: not x.startswith('_'), dir(self)))
        }


class ETLTask(Task):
    DEFAULT_CHECKPOINT = '1970-01-01 00:00:00'

    ATTR_LAST_CHECKPOINT = 'LAST_CHECKPOINT'
    ATTR_CHECKPOINT = 'CHECKPOINT'

    def __init__(self):
        Task.__init__(self)

        self.set_attribute(self.ATTR_LAST_CHECKPOINT, self.DEFAULT_CHECKPOINT)
        now_ts = int(time.time())
        init_ts = datetime_str_to_timestamp(self.DEFAULT_CHECKPOINT, tz=self.checkpoint_timezone)
        checkpoint_ts = now_ts - (now_ts - init_ts) % self.checkpoint_round
        self.set_attribute(self.ATTR_CHECKPOINT, timestamp_to_datetime(checkpoint_ts).strftime('%Y-%m-%d %H:%M:%S'))

    def on_pending(self, context):
        Task.on_pending(self, context)

        # 基于当前时间,进行粒度对齐后计算本次执行的checkpoint
        now_ts = int(time.time())
        init_ts = datetime_str_to_timestamp(self.DEFAULT_CHECKPOINT, tz=self.checkpoint_timezone)
        checkpoint_ts = now_ts - (now_ts - init_ts) % self.checkpoint_round
        self.set_attribute(self.ATTR_CHECKPOINT, timestamp_to_datetime(checkpoint_ts).strftime('%Y-%m-%d %H:%M:%S'))

        last_record = context.sys_recorder.last_success_record(self.name)
        if last_record:
            last_attrs = json.loads(last_record['attributes'])
            self.set_attribute(self.ATTR_LAST_CHECKPOINT, last_attrs[self.ATTR_CHECKPOINT])

    @property
    def _last_checkpoint(self):
        return self._attributes.get(self.ATTR_LAST_CHECKPOINT)

    @property
    def _checkpoint(self):
        return self._attributes.get(self.ATTR_CHECKPOINT)

    @property
    def checkpoint_round(self):
        """
        the time interval the checkpoint will align to
        default value is 1 day
        :return:
        """
        return 3600 * 24

    @property
    def checkpoint_timezone(self):
        """
        the timezone used when recording checkpoint
        default: None, use the local timezone
        :return:
        """
        return None

    @property
    def target_conn(self):
        """
        the target connection to write the result
        :return:
        """
        raise NotImplementedError("The target is required")

    @property
    def target_table(self):
        """
        the target table to store the result
        :return:
        """
        return self.name

    @property
    def target_mode(self):
        """
        what to do if the target table exists, replace / append / fail
        :return:
        """
        return 'replace'

    @property
    def target_typehints(self):
        """
        a dict of column_name => datatype, to customize the data type before write target
        :return:
        """
        return {}

    @property
    def target_pkey(self):
        """
        a string or a string-tuple to specify the primary key on the target table
        :return:
        """
        return None

    @property
    def target_indexes(self):
        """
        a string or a string-tuple or a string/string-tuple list to specify the indexes on the target table
        :return:
        """
        """
        :return:
        """
        return []

    @property
    def target_checkpoint_column(self):
        """
        the columns to use as the clue of last/checkpoint in target table
        :return:
        """
        return None

    def execute_internal(self, context, **kwargs):
        """
        the internal execution process to be implemented
        :param context:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def on_start(self, context, **kwargs):
        force = kwargs.get('force', False)

        if self._checkpoint > self._last_checkpoint or force:
            return self.STATE_EXECUTING
        # 重复执行就直接跳过
        logger(exec=self.flow_id, flow=self._flow, task=self.name).warn('last checkpoint {} indicates the task {} is already executed, bypass the execution'.format(
            self._last_checkpoint, self.name))

        return self.STATE_SUCCESS

    def on_commit(self, context, **kwargs):
        target_df = self._result

        enable_stdout = context.conf['pipe.stdout'] if context.conf.has('pipe.stdout') else False

        if enable_stdout and not sys.stdout.isatty():
            target_df.to_json(sys.stdout, orient='records')
            return

        target_conn = context.get_connection(self.target_conn)

        if self.target_pkey:
            assert isinstance(self.target_pkey, str) or isinstance(self.target_pkey, tuple), \
                "target primary key can only be of type string or tuple"

        if not self.target_indexes:
            indexes = []
        elif isinstance(self.target_indexes, tuple):
            indexes = [self.target_indexes]
        else:
            indexes = self.target_indexes

        for index in indexes:
            assert isinstance(index, str) or isinstance(index, tuple), \
                "target indexes can only be of type string or tuple"
        target_conn.store(target_df, self.target_table,
                          if_exists=self.target_mode,
                          chunksize=kwargs.get('chunksize', 10000),
                          typehints=self.target_typehints,
                          checkpoint=self._checkpoint,
                          last_checkpoint=self._last_checkpoint,
                          checkpoint_column=self.target_checkpoint_column,
                          pkey=self.target_pkey,
                          indexes=indexes)


class SingleSourceETLTask(ETLTask):
    @property
    def source_conn(self):
        """
        the source connection to write the result
        :return:
        """
        raise NotImplementedError("The source connection is required")

    @property
    def source(self):
        """
        the single source (table/query) to process etl
        :return:
        """
        raise NotImplementedError("The source is required")

    @property
    def is_source_query(self):
        """
        whether the source is query or not (table)
        :return:
        """
        return True

    def transform(self, df):
        """
        process to transform the source dataframe to another one
        :return:
        """
        return df

    def execute_internal(self, context, **kwargs):
        source_conn = context.get_connection(self.source_conn)
        assert isinstance(source_conn, Connection)
        enable_stdin = bool(context.conf.get_or_else('pipe.stdin', False))
        if enable_stdin:
            df = pd.read_json(sys.stdin, orient='records') if not sys.stdin.isatty() else source_conn.load_query(
                self.source) if self.is_source_query else source_conn.load(self.source)
        else:
            df = source_conn.load_query(self.source) if self.is_source_query else source_conn.load(self.source)
        return self.transform(df)


class APITask(Task):
    ATTR_TOTAL_ELEMENTS = 'totalElements'
    ATTR_VIEW_LABELS = 'labels'
    ATTR_EXPORT_LABELS = 'export_labels'

    def execute_internal(self, context, **kwargs):
        raise NotImplementedError

    def execute(self, context, **kwargs):
        raw = self.execute_internal(context, **kwargs)
        self._attributes[APITask.ATTR_VIEW_LABELS] = self.labels
        self._attributes[APITask.ATTR_EXPORT_LABELS] = self.export_labels
        return raw

    @property
    def labels(self):
        return {}

    @property
    def export_labels(self):
        return {}


class Milestone(Task):
    def __init__(self):
        Task.__init__(self)
        self._deps = []
        self._notify_success = False

    def set_deps(self, deps):
        self._deps = deps

    def enable_notify_success(self):
        self._notify_success = True

    def disable_notify_success(self):
        self._notify_success = False

    def execute_internal(self, context, **kwargs):
        pass

    @property
    def deps(self):
        raise self._deps

    @property
    def notify_success(self):
        return self._notify_success


class Flow(object):

    STATE_EXECUTING = (2, 'Executing')[0]
    STATE_SUCCESS = (3, 'Succeeded')[0]
    STATE_FAILED = (4, 'Failed')[0]

    def __init__(self, name, tasks, deps=None):
        self.name = name
        self.tasks = tasks
        self.deps = deps if deps else {}
        self.forest = self.validate()

    def __repr__(self):
        return self.name + ": tasks=" + self.tasks.__repr__() + ", deps=" + self.deps.__repr__()

    def validate(self):
        """
        util method to build flow forest
        :param tasks: the specified tasks to form the flow
        :return: the tasks are not dependent by others
        """
        assert (self.name not in self.tasks), 'flow name conflict with some tasks'

        task_roots = dict()
        for task in self.tasks:
            if task in self.deps:
                for dep in self.deps[task]:
                    task_roots[dep] = task

        forests = set(filter(lambda t: t not in task_roots, self.tasks))
        return forests

    def uniform(self):
        if len(self.forest) > 1:
            self.tasks.append(self.name)
            self.deps[self.name] = self.forest
        return self

    def dump(self, name=None, prefix='', tail=True, root=True):
        if not name:
            name = self.name
        if root:
            print(prefix + ' ' + name)
        else:
            print(prefix + ("└── " if tail else "├── ") + name)
        children = list(self.deps[name] if name in self.deps else [])

        if len(children) > 0:
            last_child = children[-1]
            children = children[0: -1]
            for child in children:
                self.dump(child, prefix + ("    " if tail else "│   "), False, False)
            self.dump(last_child, prefix + ("    " if tail else "│   "), True, False)

    def to_dict(self):
        return {
            'name': self.name,
            'tasks': self.tasks,
            'deps': dict([(t, list(d)) for (t, d) in self.deps.items()])
        }



from flask import request
from flask_restful import Api, abort, reqparse

from . import parade_blueprint, ParadeResource

api = Api(parade_blueprint, catch_all_404s=True)
parser = reqparse.RequestParser()


class FlowListAPI(ParadeResource):
    def get(self):
        flowstore = self.context.get_flowstore()
        return list(flowstore.list())

    def post(self):
        json = request.get_json()
        flowstore = self.context.get_flowstore()
        flowstore.create(json['name'], *json['tasks'], deps=json['deps'])


class FlowAPI(ParadeResource):
    """
    The api blue print to execute etl task
    """

    def get(self, flow):
        uniform = request.args.get('uniform', type=bool, default=True)
        flow_obj = self.context.get_flowstore().load(flow)
        if uniform:
            flow_obj = flow_obj.uniform()
        if not flow_obj:
            abort(404, message='Flow [{}] not found.'.format(flow))
        return flow_obj.to_dict()

    def delete(self, flow):
        flow_obj = self.context.get_flowstore().load(flow)
        if not flow_obj:
            abort(404, message='Flow [{}] not found.'.format(flow))
        self.context.get_flowstore().delete(flow)
        return '', 204


api.add_resource(FlowListAPI, '/api/flow')
api.add_resource(FlowAPI, '/api/flow/<flow>')



import os
from shutil import ignore_patterns

import parade
from . import ParadeCommand
from ..utils.misc import copytree

IGNORE = ignore_patterns('*.pyc', '.svn', '.git')


class NotebookCommand(ParadeCommand):
    requires_workspace = True

    def run_internal(self, context, **kwargs):
        NotebookCommand.init_jupyter(context)
        os.system('jupyter notebook --ip=0.0.0.0 --allow-root')

    @staticmethod
    def init_jupyter(context):
        jupyter_path = os.path.join(context.workdir, '.jupyter')
        NotebookCommand._init_jupyter(jupyter_path)
        # We need set PARADE_WORKSPACE since we'll head to notebook dir
        os.environ['PARADE_WORKSPACE'] = context.workdir
        os.chdir(os.path.join(jupyter_path, 'notebook'))

    @staticmethod
    def _init_jupyter(jupyter_path):
        if not os.path.exists(jupyter_path):
            jupyter_source_dir = os.path.join(parade.__path__[0], 'template', 'workspace', 'jupyter')
            copytree(jupyter_source_dir, jupyter_path, IGNORE)

        os.environ['JUPYTER_PATH'] = jupyter_path
        os.environ['IPYTHONDIR'] = os.path.join(jupyter_path, 'ipython')

    def short_desc(self):
        return 'start a jupyter notebook server'

    def config_parser(self, parser):
        pass



import logging
import os

logging.getLogger('werkzeug').setLevel(logging.ERROR)
logging.getLogger('socketio').setLevel(logging.ERROR)
logging.getLogger('engineio').setLevel(logging.ERROR)

_parade_formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
_console_handler = logging.StreamHandler()
_console_handler.setFormatter(_parade_formatter)


def _reset_logger(logger):
    logger.setLevel(logging.DEBUG)
    logger.handlers = [_console_handler]


parade_logger = logging.getLogger('Parade')
_reset_logger(parade_logger)


def logger(exec=None, flow=None, task=None):
    if not exec:
        _reset_logger(parade_logger)
        return parade_logger

    if task:
        parade_exec_logger = logging.getLogger('Parade.Task[{}/{}]'.format(flow, task))
        _reset_logger(parade_exec_logger)
        log_dir = os.path.join('executing', str(exec), 'tasks')
        os.makedirs(log_dir, exist_ok=True)
        handler = logging.FileHandler(os.path.join(log_dir, task))
    else:
        parade_exec_logger = logging.getLogger('Parade.Flow[{}]'.format(flow))
        _reset_logger(parade_exec_logger)
        log_dir = os.path.join('executing', str(exec))
        os.makedirs(log_dir, exist_ok=True)
        handler = logging.FileHandler(os.path.join(log_dir, flow))

    handler.setFormatter(_parade_formatter)
    parade_exec_logger.addHandler(handler)
    return parade_exec_logger



import re

date_regex = re.compile(r'''^\d{1,4}[-\/\.\s]\S+[-\/\.\s]\S+''')


def is_date(value):
    return date_regex.match(value)


def create_date_formats(day_first=True):
    """generate combinations of time and date
    formats with different delimeters
    """

    if day_first:
        date_formats = ['dd/mm/yyyy', 'dd/mm/yy', 'yyyy/mm/dd']
        python_date_formats = ['%d/%m/%Y', '%d/%m/%y', '%Y/%m/%d']
    else:
        date_formats = ['mm/dd/yyyy', 'mm/dd/yy', 'yyyy/mm/dd']
        python_date_formats = ['%m/%d/%Y', '%m/%d/%y', '%Y/%m/%d']

    date_formats += [
        # Things with words in
        'dd/bb/yyyy', 'dd/bbb/yyyy'
    ]
    python_date_formats += [
        # Things with words in
        '%d/%b/%Y', '%d/%B/%Y'
    ]

    both_date_formats = list(zip(date_formats, python_date_formats))

    # time_formats = "hh:mmz hh:mm:ssz hh:mmtzd hh:mm:sstzd".split()
    time_formats = "hh:mm:ssz hh:mm:ss hh:mm:sstzd".split()
    python_time_formats = "%H:%M%Z %H:%M:%S %H:%M:%S%Z %H:%M%z %H:%M:%S%z".split()
    both_time_formats = list(zip(time_formats, python_time_formats))

    # date_seperators = ["-","."," ","","/","\\"]
    date_seperators = ["-", ".", "/", " "]

    all_date_formats = []

    for seperator in date_seperators:
        for date_format, python_date_format in both_date_formats:
            all_date_formats.append(
                (date_format.replace("/", seperator),
                 python_date_format.replace("/", seperator))
            )

    all_formats = {}

    for date_format, python_date_format in all_date_formats:
        all_formats[date_format] = python_date_format
        for time_format, python_time_format in both_time_formats:
            all_formats[date_format + time_format] = \
                python_date_format + python_time_format

            all_formats[date_format + "T" + time_format] = \
                python_date_format + "T" + python_time_format

            all_formats[date_format + " " + time_format] = \
                python_date_format + " " + python_time_format
    return list(all_formats.values())


DATE_FORMATS = create_date_formats()



from parade.error import ParadeError

TASK_NOT_FOUND = (200, 404, 'task [{task}] not found')
DUPLICATED_TASK_EXIST = (201, 500, 'multiple tasks with name [{task}]')


class TaskNotFoundError(ParadeError):
    (code, status, message) = TASK_NOT_FOUND


class DuplicatedTaskExistError(ParadeError):
    (code, status, message) = DUPLICATED_TASK_EXIST



from ..core import Plugin


class Notifier(Plugin):
    """
    The notifier to notify task event (success/fail)
    """

    def notify_success(self, task, **kwargs):
        """
        schedule the task-flow
        :param dag_key: the flow name
        :param cron: the cron string to schedule the flow
        :return:
        """
        raise NotImplementedError

    def notify_error(self, task, reason, **kwargs):
        """
        unschedule the task-flow
        :param flow_name: the flow name
        :return:
        """
        raise NotImplementedError




import threading
from concurrent.futures import ThreadPoolExecutor

from ..core.task import Flow
from ..error.flow_errors import FlowNotFoundError
from ..utils.log import parade_logger as logger


class Engine(object):
    """
    An engine to execute etl tasks at runtime. The engine is initialized by an ETL executor context.
    """

    thread_pool = ThreadPoolExecutor(4)

    def __init__(self, context):
        """
        Initialize the Engine with the executor context
        :param context: the parade context to boot engine
        :param detach: the flag to indicate the engine is executed in detached mode or not
        :return: the initialized engine
        """
        self.context = context

    # Global lock for creating global Engine instance
    _instance_lock = threading.Lock()

    @staticmethod
    def instance(context=None, detach=False):
        """
        the static method to get or init the singleton parade-engine
        :param context: the context to init the engine
        :param detach: the flag to indicate the engine is executed in detached mode or not
        :return:
        """
        if not hasattr(Engine, "_instance"):
            with Engine._instance_lock:
                if not hasattr(Engine, "_instance"):
                    # New instance after double check
                    assert context is not None, "no context specified"
                    Engine._instance = Engine(context, detach)
        return Engine._instance

    def execute(self, task_name, **kwargs):
        """
        execute a single task with provided arguments
        :param task_name: the task name
        :param kwargs: the arguments
        :return:
        """
        task = self.context.get_task(task_name)
        task.execute(self.context, **kwargs)
        return task.result_code, task.result, task.attributes

    def execute_async(self, flow_name=None, tasks=None, new_thread=False, force=False, nodep=False):
        """
        Execute a flow or a series tasks in async mode
        :param flow_name: the flow to execute
        :param tasks: the tasks to execute
        :param new_thread: do we need a new thread to handle execution
        :param force: do we need to execute without regarding to checkpoint check
        :param nodep: do we need ignore the dependencies information in tasks. Ignore this if we execute a flow rather than tasks
        :return: tuple (exec_id, last_status) for execution id of issued flow and last status
        """
        flowstore = self.context.get_flowstore()
        # prefer to execute flow at first
        if flow_name:
            flow = flowstore.load(flow_name)
            if not flow:
                raise FlowNotFoundError(flow=flow_name)
            return self._execute_flow(flow, new_thread=new_thread, force=force)

        if len(tasks) == 0:
            tasks = list(self.context.list_tasks())
            logger.info('no task provided, use detected {} tasks in workspace {}'.format(len(tasks), tasks))

        if len(tasks) == 1:
            logger.info('single task {} provided, ignore its dependencies'.format(tasks[0]))
            nodep = True

        deps = None
        if not nodep:
            deps = dict([(task.name, task.deps) for task in self.context.load_tasks().values() if len(task.deps) > 0])
        flow = Flow(self.context.name, tasks, deps)
        logger.info('Temporary flow [{}] created!'.format(flow.name))
        return self._execute_flow(flow, new_thread=new_thread, force=force)

    def _execute_flow(self, flow, new_thread=False, force=False):
        flow_id = self.context.on_flow_start(flow)
        flowrunner = self.context.get_flowrunner()

        if new_thread:
            self.thread_pool.submit(flowrunner.submit, flow, flow_id=flow_id, force=force)
            return flow_id
        else:
            return flowrunner.submit(flow, flow_id=flow_id, force=force)



from flask_restful import Api

from . import parade_blueprint, ParadeResource, catch_parade_error
from ..error.task_errors import TaskNotFoundError

api = Api(parade_blueprint, catch_all_404s=True)


class TaskListAPI(ParadeResource):
    """
    The api blue print to execute etl task
    """

    def get(self):
        return list(self.context.list_tasks())


class TaskAPI(ParadeResource):
    """
    The api blue print to execute etl task
    """

    @catch_parade_error
    def get(self, task):
        task = self.context.get_task(task)

        return {
            'name': task.name,
            'class': type(task).__name__,
            'module': type(task).__module__,
            'bases': [x.__module__ + '.' + x.__name__ for x in type(task).__bases__],
            # 'attrs': list(filter(lambda x: x not in dir(Task) and not x.startswith('_'), dir(task)))
            'attrs': list(filter(lambda x: not x.startswith('_'), dir(task)))
        }


api.add_resource(TaskListAPI, '/api/task')
api.add_resource(TaskAPI, '/api/task/<task>')



"""Helper functions which don't fit anywhere else"""
import os

from shutil import copy2, copystat


def copytree(src, dst, ignore=None):
    """
    Since the original function always creates the directory, to resolve
    the issue a new function had to be created. It's a simple copy and
    was reduced for this case.

    More info at:
    https://github.com/scrapy/scrapy/pull/2005
    """
    names = os.listdir(src)
    ignored_names = ignore(src, names)

    if not os.path.exists(dst):
        os.makedirs(dst)

    for name in names:
        if name in ignored_names:
            continue

        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        if os.path.isdir(srcname):
            copytree(srcname, dstname, ignore)
        else:
            copy2(srcname, dstname)
    copystat(src, dst)



from ...core.context import Context
from .. import ParadeCommand
from ...utils.log import parade_logger as logger


class MakeFlowCommand(ParadeCommand):

    requires_workspace = True

    @property
    def aliases(self):
        return ['new']

    def run_internal(self, context, **kwargs):
        assert isinstance(context, Context)
        deps = dict()
        flow_and_task = kwargs.get('flow_and_task')
        if len(flow_and_task) == 0:
            flow = context.name
            tasks = context.list_tasks()
            logger.info('no task provided, use detected {} tasks in workspace {}'.format(len(tasks), tasks))
            deps = dict([(task.name, task.deps) for task in context.load_tasks().values() if len(task.deps) > 0])
        else:
            flow = flow_and_task[0]
            tasks = flow_and_task[1:]
            assert all(t in context.list_tasks() for t in tasks), 'some task in {} not found in workspace'.format(tasks)
            _deps = kwargs.get('dep')
            if _deps:
                for dep in _deps:
                    (l, r) = tuple(dep.split('->'))
                    if l not in tasks:
                        logger.error('task {} in dependencies not specified'.format(l))
                    if r not in tasks:
                        logger.error('task {} in dependencies not specified'.format(r))
                    deps[l] = deps[l] if l in deps else set(r)
                    assert isinstance(deps[l], set)
                    deps[l].add(r)

        flowstore = context.get_flowstore()
        flowstore.create(flow, *tasks, deps=deps)
        print('Flow {} created, details:'.format(flow))
        flow = flowstore.load(flow).uniform()
        print('tasks: {}'.format(flow.tasks))
        print('dependencies:')
        print('------------------------------------------')
        flow.dump()
        print('------------------------------------------')

    def short_desc(self):
        return 'create a dag (flow) with a set of tasks'

    def config_parser(self, parser):
        parser.add_argument('flow_and_task', nargs='*', help='the flow name and tasks')
        parser.add_argument('-d', '--dep', action='append')




import sys
import pandas as pd
from ..core.task import Task
from ..utils.log import parade_logger as logger

from . import ParadeCommand


class ConvertCommand(ParadeCommand):
    """
    The install command will install acontrib component into current workspace
    """

    requires_workspace = True

    def short_desc(self):
        return 'convert a source table to a target table'

    def config_parser(self, parser):
        parser.add_argument('-s', '--source', dest='source', help='the source table [conn_key:table_name]')
        parser.add_argument('-t', '--target', dest='target', help='set target table [conn_key:table_name]')

    def run_internal(self, context, **kwargs):
        source = kwargs.get('source')
        target = kwargs.get('target')

        if not sys.stdin.isatty() or not source:
            logger.info('use stdin as source')
            df = pd.read_json(sys.stdin, orient='records')
        else:
            if ':' not in source:
                logger.error('source should be in format [conn_key:table_name]')
                return Task.RET_CODE_FAIL
            else:
                (source_conn, source_table) = tuple(source.split(':'))
                df = context.get_connection(source_conn).load(source_table)

        if not sys.stdout.isatty() or not target:
            logger.info('use stdout as source')
            df.to_json(sys.stdout, orient='records')
        else:
            if ':' not in target:
                logger.error('target should be in format [conn_key:table_name]')
                return Task.RET_CODE_FAIL
            else:
                (target_conn, target_table) = tuple(target.split(':'))
                context.get_connection(target_conn).store(df, target_table)

        return Task.RET_CODE_SUCCESS




from parade.command import ParadeCommand
from parade.core.engine import Engine


class ExecCommand(ParadeCommand):
    """
    The exec command to run a flow or a set tasks,
    if the tasks to execute have dependencies on each other,
    parade will handle them correctly
    """
    requires_workspace = True

    @property
    def aliases(self):
        return ['execute', 'exec']

    def run_internal(self, context, **kwargs):
        engine = Engine(context)

        tasks = kwargs.get('task')
        force = kwargs.get('force')
        nodep = kwargs.get('nodep')

        return engine.execute_async(tasks=tasks, force=force, nodep=nodep)

    def short_desc(self):
        return 'execute a set of tasks'

    def config_parser(self, parser):
        parser.add_argument('--nodep', action="store_true", help='execute tasks without considering dependencies')
        parser.add_argument('--force', action="store_true", help='force the task to execute')
        parser.add_argument('task', nargs='*', help='the task to schedule')



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from setuptools import setup, find_packages

setup(
    name='parade',
    version='0.1.21.0',
    description='A ETL engine for pandas-based data task',
    author='He Bai',
    author_email='bailaohe@gmail.com',
    packages=find_packages('src'),
    package_dir=({'parade': 'src/parade'}),
    zip_safe=False,
    include_package_data=True,
    package_data={'': ['*.json', '*.xml', '*.yml', '*.tpl']},
    install_requires=['pandas', 'sqlalchemy', 'requests', 'pyyaml', 'flask',
                      'flask_cors', 'flask_restful', 'psycopg2', 'mysqlclient', 'xlsxwriter',
                      'PyGithub', 'pymongo', 'Flask-SocketIO', 'redis', 'jupyter', 'nbformat'],
    entry_points={
        'console_scripts': [
            'parade = parade.cmdline:execute',
        ],
    },
    url='https://github.com/bailaohe/parade',
    download_url='https://github.com/bailaohe/parade/tarball/0.1',
)



from ..core import Plugin


class FlowRunner(Plugin):
    """
    The executor to submit task flow
    """
    def submit(self, flow, flow_id=0, **kwargs):
        """
        schedule the task-flow
        :param flow: the flow to run
        :return:
        """
        raise NotImplementedError




import copy

from ..utils.modutils import get_class
from ..core import Plugin


class ConfigEntry(object):
    def __init__(self, value=None, prefix=None):
        if not value:
            value = {}
        if not isinstance(value, dict):
            raise TypeError('Can only accept dict')
        self._dict = value
        self.prefix = prefix

    def put(self, path, value):
        assert path is not None
        assert value is not None
        assert isinstance(path, str)
        # assert isinstance(value, str)

        paths = path.split('.')

        node = self._dict
        for i in range(len(paths) - 1):
            if paths[i] not in node:
                node[paths[i]] = {}
            node = node[paths[i]]
        node[paths[-1]] = value

    def get_or_else(self, path=None, default=None):
        try:
            return self.get(path)
        except RuntimeError:
            return default

    def get(self, path=None):

        if not path:
            return self._dict

        tokens = path.split(".")

        conf_node = self._dict
        for item in tokens:
            if item not in conf_node:
                raise RuntimeError('load config [{}] failed'.format(path))
            conf_node = conf_node[item]

        if isinstance(conf_node, dict):
            return ConfigEntry(conf_node, path if not self.prefix else self.prefix + '.' + path)
        return conf_node

    def has(self, path=None):
        try:
            return self.get(path) is not None
        except RuntimeError:
            return False

    def traverse(self):
        stack = [([], self._dict)]
        results = {}
        while len(stack) > 0:
            keys, val = stack.pop()

            if isinstance(val, dict):
                for (k, v) in val.items():
                    stack.append((keys + [k], v))

            else:
                results['.'.join(keys)] = val

        return results

    def to_dict(self):
        return self._dict

    def __getitem__(self, path):
        return self.get(path)

    def __str__(self):
        return self._dict.__str__()


class ConfigParser(object):
    def parse(self, raw_conf, **kwargs):
        raise NotImplementedError


class ConfigStore(Plugin):
    def load(self, **kwargs):
        import os
        config_name = self.conf['name']
        config_profile = self.conf['profile']

        config_profile = os.environ.get('PARADE_PROFILE', config_profile)

        raw_conf = self.load_internal(config_name, profile=config_profile)
        conf = copy.deepcopy(raw_conf)
        if self.conf.has('parser'):
            parser_driver = self.conf['parser']
            parser_cls = get_class(parser_driver, ConfigParser, self.context.name + '.contrib.config')
            config_parser = parser_cls()
            conf = config_parser.parse(raw_conf, profile=config_profile)

        return ConfigEntry(conf)

    def load_internal(self, name, profile='default', **kwargs):
        raise NotImplementedError



from flask import request
from flask_restful import Api, reqparse

from . import parade_blueprint, ParadeResource, catch_parade_error
from ..core.engine import Engine

api = Api(parade_blueprint, catch_all_404s=True)
parser = reqparse.RequestParser()


class ExecAPI(ParadeResource):
    @catch_parade_error
    def post(self):
        parser.add_argument('flow', type=str)
        parser.add_argument('task', action='append')
        parser.add_argument('force', type=bool)
        parser.add_argument('nodep', type=bool)

        args = parser.parse_args()
        flow = args.get('flow', None)
        tasks = args.get('task', [])
        force = args.get('force', False)
        nodep = args.get('nodep', False)

        engine = Engine(self.context)
        return engine.execute_async(flow, tasks, new_thread=True, force=force, nodep=nodep)

    @catch_parade_error
    def get(self):
        # parser.add_argument('executing', type=bool)
        # args = parser.parse_args()
        executing = request.args.get('executing', type=bool, default=None)
        flow = request.args.get('flow', default=None)
        page_size = request.args.get('pageSize', type=int, default=0)
        page_no = request.args.get('pageNo', type=int, default=1)
        return self.context.sys_recorder.load_flows(flow=flow, executing=executing, page_size=page_size,
                                                    page_no=page_no)


class ExecDetailAPI(ParadeResource):
    @catch_parade_error
    def get(self, id):
        executing_flow = self.context.sys_recorder.load_flow_by_id(id)
        exec_tasks = self.context.sys_recorder.load_flow_tasks(id)
        flow = self.context.get_flowstore().load(executing_flow['flow']).uniform()

        result = {
            'flow': flow.to_dict(),
            'exec': {
                'flow': executing_flow,
                'tasks': exec_tasks
            }
        }
        return result


class JobLogAPI(ParadeResource):
    @catch_parade_error
    def get(self, exec_id, task):
        import os
        logfile = os.path.join('executing', exec_id, 'tasks', task)

        log_lines = []
        if os.path.exists(logfile):
            with open(logfile, 'r') as f:
                log_lines = f.readlines()

        return log_lines


api.add_resource(ExecAPI, '/api/exec')
api.add_resource(ExecDetailAPI, '/api/exec/<id>')
api.add_resource(JobLogAPI, '/api/exec/<exec_id>/<task>')



from ..core import Plugin


class Datasource(object):
    """
    The data source object. The object does not maintain any stateful information.
    """

    def __init__(self, driver, **kwargs):
        self.driver = driver
        self.attributes = kwargs

    @property
    def protocol(self):
        return self.attributes['protocol'] if 'protocol' in self.attributes else None

    @property
    def host(self):
        return self.attributes['host'] if 'host' in self.attributes else None

    @property
    def port(self):
        return self.attributes['port'] if 'port' in self.attributes else None

    @property
    def user(self):
        return self.attributes['user'] if 'user' in self.attributes else None

    @property
    def password(self):
        return self.attributes['password'] if 'password' in self.attributes else None

    @property
    def db(self):
        return self.attributes['db'] if 'db' in self.attributes else None

    @property
    def uri(self):
        return self.attributes['uri'] if 'uri' in self.attributes else None


class Connection(Plugin):
    """
    The connection object, which is opened with data source and its implementation is also
    related to the context
    """
    datasource = None

    def initialize(self, context, conf):
        Plugin.initialize(self, context, conf)
        self.datasource = Datasource(**conf.to_dict())

    def load(self, table, **kwargs):
        raise NotImplementedError

    def load_query(self, query, **kwargs):
        raise NotImplementedError

    def store(self, df, table, **kwargs):
        raise NotImplementedError



"""The Parade data processing engine and services."""

from __future__ import absolute_import, division, print_function, with_statement

# version is a human-readable version number.
version = "0.1.0"

# version_info is a four-tuple for programmatic comparison. The first
# three numbers are the components of the version number.  The fourth
# is zero for an official release, positive for a development branch,
# or negative for a release candidate or beta (after the base version
# number has been incremented)
version_info = (0, 1, 0, -1)



from ..core.context import Context
from ..utils.workspace import inside_workspace


class ParadeCommand(object):
    """
    The base class to implement a sub-command in parade command line entry point
    """
    requires_workspace = False
    workspace_name = None
    workspace_path = None
    config = None

    @property
    def aliases(self):
        return []

    def short_desc(self):
        """
        A short description of the command
        """
        raise NotImplementedError

    def long_desc(self):
        """A long description of the command. Return short description when not
        available. It cannot contain newlines, since contents will be formatted
        by optparser which removes newlines and wraps text.
        """
        return self.short_desc()

    def help(self):
        """An extensive help for the command. It will be shown when using the
        "help" command. It can contain newlines, since not post-formatting will
        be applied to its contents.
        """
        return self.long_desc()

    def config_parser(self, parser):
        """
        Populate option parse with options available for this command
        """
        raise NotImplementedError

    def run(self, context, **kwargs):
        """
        Entry point for running commands
        """
        return self.run_internal(context, **kwargs)

    def run_internal(self, context, **kwargs):
        """
        Internal running logic, to be implemented
        :param context: the context to use
        :return:
        """
        raise NotImplementedError



from .. import ParadeCommand


class ListFlowCommand(ParadeCommand):

    requires_workspace = True

    @property
    def aliases(self):
        return ['ls']

    def run_internal(self, context, **kwargs):
        flowstore = context.get_flowstore()
        flows = flowstore.list()
        print('Workspace [{}] has {} flow(s):'.format(context.name, len(flows)))
        for flow_id, flow_name in enumerate(flows):
            print("#{}: {}".format(flow_id + 1, flow_name))

    def short_desc(self):
        return 'list the flows created'

    def config_parser(self, parser):
        pass

